# Modified version of robin_stocks login to enable asynchronous flows
# such as in a web server

import os
import pickle

import platformdirs
from pathlib import Path
from py_portfolio_index.exceptions import (
    ExtraAuthenticationStepException,
    ConfigurationError,
)
from py_portfolio_index.models import LoginResponse, LoginResponseStatus

ROBINHOOD_USERNAME_ENV = "VANGUARD_USERNAME"
ROBINHOOD_PASSWORD_ENV = "VANGUARD_PASSWORD"


def validate_login():
    from robin_stocks.robinhood.authentication import (
        request_get,
        positions_url,
    )

    res = request_get(
        positions_url(),
        "pagination",
        {"nonzero": "true"},
        jsonify_data=False,
    )
    # Raises exception is response code is not 200.
    try:
        res.raise_for_status()
    except Exception:
        raise ConfigurationError()


def login(
    username=None,
    password=None,
    expiresIn=86400,
    scope="internal",
    by_sms=True,
    store_session=True,
    challenge_response: str | int | None = None,
    pickle_name="",
    prior_response=LoginResponse | None,
):
    from robin_stocks.robinhood.authentication import (
        generate_device_token,
        update_session,
        set_login_state,
        respond_to_challenge,
        request_get,
        positions_url,
        request_post,
    )

    """This function will effectively log the user into robinhood by getting an
    authentication token and saving it to the session header. By default, it
    will store the authentication token in a pickle file and load that value
    on subsequent logins.

    :param username: The username for your robinhood account, usually your email.
        Not required if credentials are already cached and valid.
    :type username: Optional[str]
    :param password: The password for your robinhood account. Not required if
        credentials are already cached and valid.
    :type password: Optional[str]
    :param expiresIn: The time until your login session expires. This is in seconds.
    :type expiresIn: Optional[int]
    :param scope: Specifies the scope of the authentication.
    :type scope: Optional[str]
    :param by_sms: Specifies whether to send an email(False) or an sms(True)
    :type by_sms: Optional[boolean]
    :param store_session: Specifies whether to save the log in authorization
        for future log ins.
    :type store_session: Optional[boolean]
    :param mfa_code: MFA token if enabled.
    :type mfa_code: Optional[str]
    :param pickle_name: Allows users to name Pickle token file in order to switch
        between different accounts without having to re-login every time.
    :returns:  A dictionary with log in information. The 'access_token' keyword contains the access token, and the 'detail' keyword \
    contains information on whether the access token was generated or loaded from pickle file.

    """
    if not username:
        username = os.environ.get(ROBINHOOD_USERNAME_ENV, None)
    if not password:
        password = os.environ.get(ROBINHOOD_PASSWORD_ENV, None)
    device_token = generate_device_token()
    data_dir = Path(
        platformdirs.user_data_dir("py_portfolio_index", ensure_exists=True)
    )
    data_dir = data_dir / ".tokens"
    creds_file = "vanguard" + pickle_name + ".pickle"
    pickle_path = data_dir / creds_file
    if not os.path.exists(data_dir):
        os.makedirs(data_dir, exist_ok=True)
    # Challenge type is used if not logging in with two-factor authentication.
    if by_sms:
        challenge_type = "sms"
    else:
        challenge_type = "email"
    url = ""
    payload = {
        "client_id": "c82SH0WZOsabOXGP2sxqcj34FxkvfnWRZBKlBjFS",
        "expires_in": expiresIn,
        "grant_type": "password",
        "password": password,
        "scope": scope,
        "username": username,
        "challenge_type": challenge_type,
        "device_token": device_token,
    }

    # If authentication has been stored in pickle file then load it. Stops login server from being pinged so much.
    if os.path.isfile(pickle_path):
        # If store_session has been set to false then delete the pickle file, otherwise try to load it.
        # Loading pickle file will fail if the access_token has expired.
        if store_session:
            try:
                with open(pickle_path, "rb") as f:
                    pickle_data = pickle.load(f)
                    access_token = pickle_data["access_token"]
                    token_type = pickle_data["token_type"]
                    refresh_token = pickle_data["refresh_token"]
                    # Set device_token to be the original device token when first logged in.
                    pickle_device_token = pickle_data["device_token"]
                    payload["device_token"] = pickle_device_token
                    # Set login status to True in order to try and get account info.
                    set_login_state(True)
                    update_session(
                        "Authorization", "{0} {1}".format(token_type, access_token)
                    )
                    # Try to load account profile to check that authorization token is still valid.
                    res = request_get(
                        positions_url(),
                        "pagination",
                        {"nonzero": "true"},
                        jsonify_data=False,
                    )
                    # Raises exception is response code is not 200.
                    res.raise_for_status()
                    return {
                        "access_token": access_token,
                        "token_type": token_type,
                        "expires_in": expiresIn,
                        "scope": scope,
                        "detail": "logged in using authentication in {0}".format(
                            creds_file
                        ),
                        "backup_code": None,
                        "refresh_token": refresh_token,
                    }
            except Exception as e:
                print(
                    f"ERROR: There was an issue loading pickle file {str(e)}. Authentication may be expired - logging in normally."
                )
                set_login_state(False)
                update_session("Authorization", None)
                # raise ConfigurationError()
        else:
            os.remove(pickle_path)

    # Handle case where mfa or challenge is required.
    if prior_response:
        if prior_response.status == LoginResponseStatus.MFA_REQUIRED:
            assert challenge_response is not None
            payload["mfa_code"] = challenge_response
        if prior_response.status == LoginResponseStatus.CHALLENGE_REQUIRED:
            res = respond_to_challenge(
                prior_response.data["challenge_id"], challenge_response
            )
            update_session(
                "X-ROBINHOOD-CHALLENGE-RESPONSE-ID", prior_response.data["challenge_id"]
            )
    data = request_post(url, payload)

    if data:
        if "mfa_required" in data:
            raise ExtraAuthenticationStepException(
                response=LoginResponse(status=LoginResponseStatus.MFA_REQUIRED)
            )
        elif "challenge" in data:
            challenge_id = data["challenge"]["id"]
            raise ExtraAuthenticationStepException(
                response=LoginResponse(
                    status=LoginResponseStatus.CHALLENGE_REQUIRED,
                    data={"challenge_id": challenge_id},
                )
            )
        # Update Session data with authorization or raise exception with the information present in data.
        if "access_token" in data:
            token = "{0} {1}".format(data["token_type"], data["access_token"])
            update_session("Authorization", token)
            set_login_state(True)
            data["detail"] = "logged in with brand new authentication code."
            if store_session:
                with open(pickle_path, "wb") as f:
                    pickle.dump(
                        {
                            "token_type": data["token_type"],
                            "access_token": data["access_token"],
                            "refresh_token": data["refresh_token"],
                            "device_token": payload["device_token"],
                        },
                        f,
                    )
        else:
            raise Exception(data["detail"])
    else:
        raise Exception(
            "Error: Trouble connecting to vanguard API. Check internet connection."
        )
    return data



from py_portfolio_index import INDEXES, STOCK_LISTS


def test_indexes():
    for key in INDEXES.keys:
        x = INDEXES[key]

        print(x.json())
    print(INDEXES.json())


def test_lists():
    for key in STOCK_LISTS.keys:
        STOCK_LISTS[key]

    print(STOCK_LISTS.json())



from py_portfolio_index import PaperAlpacaProvider
from datetime import date


def test_provider_methods():
    providers = [PaperAlpacaProvider()]
    # if "robinhood" in AVAILABLE_PROVIDERS:
    #     providers.append(RobinhoodProvider())
    test_date = date(year=2023, month=1, day=1)
    for provider in providers:
        provider.get_holdings()
        real_ticker = "MSFT"
        p1 = provider.get_instrument_price(real_ticker)
        p2 = provider.get_instrument_prices([real_ticker])
        assert p1 == p2[real_ticker]

        p1 = provider.get_instrument_price(real_ticker, at_day=test_date)
        p2 = provider.get_instrument_prices([real_ticker], at_day=test_date)
        assert p1 == p2[real_ticker]

        provider.get_unsettled_instruments()

        provider.get_stock_info(real_ticker)


# def test_stock_reweighting():
#     provider = PaperAlpacaProvider()

#     portfolio = IdealPortfolio(source_date="2021-01-01")
#     por



from py_portfolio_index.bin import INDEXES, STOCK_LISTS
from py_portfolio_index.constants import Logger
from py_portfolio_index.enums import PurchaseStrategy, RoundingStrategy
from py_portfolio_index.operators import (
    compare_portfolios,
    generate_order_plan,
    generate_composite_order_plan,
)
from py_portfolio_index.portfolio_providers.robinhood import RobinhoodProvider
from py_portfolio_index.portfolio_providers.alpaca_v2 import (
    AlpacaProvider,
    PaperAlpacaProvider,
)
from py_portfolio_index.portfolio_providers.webull import (
    WebullProvider,
    WebullPaperProvider,
)
from py_portfolio_index.portfolio_providers.moomoo import MooMooProvider
from py_portfolio_index.config import get_providers
from py_portfolio_index.models import IdealPortfolio

AVAILABLE_PROVIDERS = get_providers()

__version__ = "0.0.32"

__all__ = [
    "INDEXES",
    "STOCK_LISTS",
    "Logger",
    "compare_portfolios",
    "generate_order_plan",
    "generate_composite_order_plan",
    "PaperAlpacaProvider",
    "AlpacaProvider",
    "WebullProvider",
    "WebullPaperProvider",
    "MooMooProvider",
    "RobinhoodProvider",
    "PurchaseStrategy",
    "RoundingStrategy",
    "IdealPortfolio",
]



from py_portfolio_index.models import RealPortfolio, RealPortfolioElement, Money


def test_serialization():
    test_port = RealPortfolio(
        holdings=[
            RealPortfolioElement(ticker="AAPL", units=1.0, value=Money(value=100))
        ]
    )

    x = test_port.json()

    print(x)



from decimal import Decimal
from datetime import date, datetime
from typing import Optional, List, Dict, DefaultDict, Any
from py_portfolio_index.models import RealPortfolio, RealPortfolioElement, Money
from py_portfolio_index.common import divide_into_batches
from py_portfolio_index.portfolio_providers.common import PriceCache
from py_portfolio_index.portfolio_providers.base_portfolio import (
    BaseProvider,
    CacheKey,
)
from py_portfolio_index.exceptions import (
    ConfigurationError,
    PriceFetchError,
    OrderError,
)
from py_portfolio_index.enums import Provider
from functools import lru_cache
from os import environ
from collections import defaultdict

FRACTIONAL_SLEEP = 60
BATCH_SIZE = 50

CACHE_PATH = "moo_moo_tickers.json"


def nearest_value(all_historicals, pivot) -> Optional[dict]:
    filtered = [z for z in all_historicals if z]
    if not filtered:
        return None
    return min(
        filtered,
        key=lambda x: abs(
            datetime.strptime(x["begins_at"], "%Y-%m-%dT%H:%M:%SZ").date() - pivot
        ),
    )


def nearest_multi_value(
    symbol: str, all_historicals, pivot: Optional[date] = None
) -> Optional[Decimal]:
    filtered = [z for z in all_historicals if z and z["symbol"] == symbol]
    if not filtered:
        return None
    if pivot is not None:
        lpivot = pivot or date.today()
        closest = min(
            filtered,
            key=lambda x: abs(
                datetime.strptime(x["begins_at"], "%Y-%m-%dT%H:%M:%SZ").date() - lpivot
            ),
        )
    else:
        closest = filtered[0]
    if closest:
        value = closest.get("last_trade_price", closest.get("high_price", None))
        return Decimal(value)
    return None


class InstrumentDict(dict):
    def __init__(self, refresher, *args):
        super().__init__(*args)
        self.refresher = refresher

    def __missing__(self, key):
        mapping = self.refresher()
        self.update(mapping)
        if key in self:
            return self[key]
        raise ValueError(f"Could not find instrument {key} after refresh")


class MooMooProvider(BaseProvider):
    """Provider for interacting with stocks held in
    MooMoo
    """

    PROVIDER = Provider.MOOMOO
    SUPPORTS_BATCH_HISTORY = 0
    PASSWORD_ENV = "MOOMOO_PASSWORD"
    USERNAME_ENV = "MOOMOO_USERNAME"
    TRADE_TOKEN_ENV = "MOOMOO_TRADE_TOKEN"
    DEVICE_ID_ENV = "MOOMOO_DEVICE_ID"

    def __init__(
        self,
        skip_cache: bool = False,
    ):
        # if not username:
        #     username = environ.get(self.USERNAME_ENV, None)
        # if not password:
        #     password = environ.get(self.PASSWORD_ENV, None)
        # if not trade_token:
        #     trade_token = environ.get(self.TRADE_TOKEN_ENV, None)
        # if not device_id:
        #     device_id = environ.get(self.DEVICE_ID_ENV, None)
        # if not (username and password and trade_token and device_id):
        #     raise ConfigurationError(
        #         "Must provide ALL OF username, password, trade_token, and device_id arguments or set environment variables MOOMOO_USERNAME, MOOMOO_PASSWORD, MOOMOO_TRADE_TOKEN, and MOOMOO_DEVICE_ID "
        #     )
        from moomoo import (
            OpenSecTradeContext,
            OpenQuoteContext,
            SecurityFirm,
            TrdMarket,
        )

        self._trade_provider = OpenSecTradeContext(
            filter_trdmarket=TrdMarket.US,
            host="localhost",
            port=11111,
            security_firm=SecurityFirm.FUTUINC,
        )
        self._quote_provider = OpenQuoteContext(host="localhost", port=11111)
        BaseProvider.__init__(self)
        self._local_latest_price_cache: Dict[str, Decimal | None] = defaultdict(
            lambda: None
        )
        self._price_cache: PriceCache = PriceCache(fetcher=self._get_instrument_prices)
        self._local_instrument_cache: Dict[str, str] = {}
        if not skip_cache:
            self._load_local_instrument_cache()

    def _load_local_instrument_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir("py_portfolio_index", ensure_exists=True))
        file = path / CACHE_PATH
        if not file.exists():
            self._local_instrument_cache = {}
            return
        with open(file, "r") as f:
            self._local_instrument_cache = json.load(f)
            # corruption guard
            if not isinstance(self._local_instrument_cache, dict):
                self._local_instrument_cache = {}

    def _save_local_instrument_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir("py_portfolio_index", ensure_exists=True))
        file = path / CACHE_PATH
        with open(file, "w") as f:
            json.dump(self._local_instrument_cache, f)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        # TODO: determine if there is a bulk API
        from moomoo import RET_OK, SubType

        if at_day:
            raise ValueError
            # historicals = self._provider.get_bars(
            #     tId=webull_id,
            #     interval="d1",
            #     timeStamp=int(
            #         datetime(
            #             day=at_day.day,
            #             month=at_day.month,
            #             year=at_day.year,
            #             tzinfo=UTC,
            #         ).timestamp()
            #     ),
            # )
            # return Decimal(value=list(historicals.itertuples())[0].vwap)
        else:
            ret_sub, err_message = self._quote_provider.subscribe(
                ["US." + ticker], [SubType.TICKER], subscribe_push=False
            )
            # Subscribe to the K line type first. After the subscription is successful, moomoo OpenD will continue to receive pushes from the server, False means that there is no need to push to the script temporarily
            if ret_sub == RET_OK:  # Subscription successful
                ret, data = self._quote_provider.get_stock_quote(
                    ["US." + ticker]
                )  # Get real-time data of subscription stock quotes
                if ret == RET_OK:
                    return list(data.itertuples())[0]
            else:
                print("subscription failed", err_message)
            raise PriceFetchError("Could not get price")

    def _buy_instrument(
        self,
        symbol: str,
        qty: Optional[float],
        value: Optional[Money] = None,
        price: Optional[Decimal] = None,
    ) -> bool:
        from moomoo import RET_OK, TrdSide, OrderType

        ret, data = self._trade_provider.unlock_trade(
            environ.get(self.TRADE_TOKEN_ENV, None)
        )  # If you use a live trading account to place an order, you need to unlock the account first. The example here is to place an order on a paper trading account, and unlocking is not necessary.
        if ret == RET_OK:
            print("unlocked orders")
        else:
            raise OrderError("unlock trade error: ", data)
        ret, data = self._trade_provider.place_order(
            price=price,
            qty=qty,
            code="US." + symbol,
            order_type=OrderType.MARKET,
            trd_side=TrdSide.BUY,
        )
        if ret == RET_OK:
            return True
        else:
            raise OrderError("place_order error: ", data)

    def buy_instrument(
        self, ticker: str, qty: Decimal, value: Optional[Money] = None
    ) -> bool:
        # TODO: make sure this is always set
        if not value and qty:
            raise NotImplementedError(
                "Moomoo provider must have both value and qty to purchase"
            )
        assert value
        price = value / qty
        if qty:
            orders_kwargs_list: List[Dict[str, Money | None | Decimal]] = [
                {"qty": qty, "value": None, "price": price}
            ]
        else:
            orders_kwargs_list = [{"qty": None, "value": value, "price": price}]
        for order_kwargs in orders_kwargs_list:
            return self._buy_instrument(ticker, **order_kwargs)  # type: ignore
        return True

    def get_unsettled_instruments(self) -> set[str]:
        """We need to efficiently bypass
        paginating all orders if possible
        so just check the account info for if there
        is any cash held for orders first"""
        from moomoo import RET_OK
        from moomoo import OrderStatus

        ret, data = self._trade_provider.order_list_query(
            status_filter_list=[
                OrderStatus.SUBMITTED,
                OrderStatus.FILLED_PART,
                OrderStatus.WAITING_SUBMIT,
            ]
        )
        if ret == RET_OK:
            print(data.head(10))
            pass
        else:
            raise ConfigurationError("Could not get order list")
        # code is of format US.MSFT, for example
        return set(item.code.split(".")[-1] for item in data.itertuples())

    def _get_stock_info(self, ticker: str) -> dict:
        raise NotImplementedError()
        # info = self._provider.get_ticker_info(ticker)
        # matches = self._provider.find_instrument_data(ticker)
        # for match in matches:
        #     if match["symbol"] == ticker:
        #         return {
        #             "name": match["simple_name"],
        #             "exchange": match["exchange"],
        #             "market": match["market"],
        #             "country": match["country"],
        #             "tradable": bool(match["tradable"]),
        #         }
        # return info

    def _get_portfolio(self):
        from moomoo import RET_OK

        ret, data = self._trade_provider.accinfo_query()
        if ret == RET_OK:
            return list(data.itertuples())[0]

        raise ConfigurationError("Could not get portfolio")

    def _get_positions(self):
        from moomoo import RET_OK

        ret, data = self._trade_provider.position_list_query()
        if ret == RET_OK:
            return data.itertuples()

        raise ConfigurationError("Could not get positions")

    def get_holdings(self) -> RealPortfolio:
        accounts_data = self._get_cached_value(
            CacheKey.ACCOUNT, callable=self._get_portfolio
        )
        my_stocks = self._get_cached_value(
            CacheKey.POSITIONS, callable=self._get_positions
        )
        unsettled = self._get_cached_value(
            CacheKey.UNSETTLED, callable=self.get_unsettled_instruments
        )

        pre = {}
        symbols = []
        total_value = Decimal(0.0)
        for row in my_stocks:
            print(row)
            local: Dict[str, Any] = {}
            local["units"] = row.qty
            # instrument_data = self._provider.get_instrument_by_url(row["instrument"])
            ticker = row.code.split(".")[-1]
            local["ticker"] = ticker
            symbols.append(ticker)
            local["value"] = Decimal(row.market_val)
            local["weight"] = 0
            pre[ticker] = local
            total_value += local["value"]

        final = []
        for s in symbols:
            local = pre[s]
            local["weight"] = local["value"] / total_value
            local["unsettled"] = s in unsettled
            final.append(local)
        out = [RealPortfolioElement(**row) for row in final]
        cash = Decimal(accounts_data.net_cash_power)
        return RealPortfolio(holdings=out, cash=Money(value=cash), provider=self)

    def get_instrument_prices(self, tickers: List[str], at_day: Optional[date] = None):
        return self._price_cache.get_prices(tickers=tickers, date=at_day)

    def _get_instrument_prices(
        self, tickers: List[str], at_day: Optional[date] = None
    ) -> Dict[str, Optional[Decimal]]:
        batches: List[Dict[str, Optional[Decimal]]] = []
        for list_batch in divide_into_batches(tickers, 1):
            # TODO: determine if there is a bulk API
            ticker: str = list_batch[0]
            # webull_id = self._local_instrument_cache.get(ticker)
            rval = self._get_instrument_price(ticker, at_day=at_day)
            batches.append({ticker: rval})
        prices: Dict[str, Optional[Decimal]] = {}
        for fbatch in batches:
            prices = {**prices, **fbatch}
        return prices

    def get_profit_or_loss(self, include_dividends: bool = True) -> Money:
        raise NotImplementedError()
        # my_stocks = self._get_cached_value(
        #     CacheKey.POSITIONS, callable=self._provider.get_positions
        # )
        # pls: List[Money] = []
        # for x in my_stocks:
        #     pl = Money(value=Decimal(x["unrealizedProfitLoss"]))
        #     pls.append(pl)
        # _total_pl = sum(pls)  # type: ignore
        # if not include_dividends:
        #     return Money(value=_total_pl)
        # return Money(value=_total_pl) + sum(self._get_dividends().values())

    def _get_dividends(self) -> DefaultDict[str, Money]:
        # dividends = self._provider.get_dividends()
        out: DefaultDict[str, Money] = DefaultDict(lambda: Money(value=0))
        return out


# class MooMooPaperProvider(MooMooProvider):
#     PROVIDER = Provider.MOOMOO_PAPER
#     PASSWORD_ENV = "MOOMOO_PAPER_PASSWORD"
#     USERNAME_ENV = "MOOMOO_PAPER_USERNAME"
#     TRADE_TOKEN_ENV = "MOOMOO_PAPER_TRADE_TOKEN"
#     DEVICE_ID_ENV = "MOOMOO_PAPER_DEVICE_ID"

#     def _get_provider(self):
#         from webull import paper_webull

#         return paper_webull



from dataclasses import dataclass
from typing import Optional, Dict, Union, Mapping, List
from decimal import Decimal
from math import floor, ceil
from collections import defaultdict

from py_portfolio_index.common import print_per
from py_portfolio_index.constants import Logger
from py_portfolio_index.enums import PurchaseStrategy, RoundingStrategy
from py_portfolio_index.portfolio_providers.base_portfolio import BaseProvider
from py_portfolio_index.models import (
    Money,
    Provider,
    OrderElement,
    OrderPlan,
    OrderType,
    PortfolioProtocol,
    CompositePortfolio,
)
from .models import IdealPortfolio

MIN_ORDER_SIZE = 2
MIN_ORDER_MONEY = Money(value=MIN_ORDER_SIZE)


@dataclass
class ComparisonResult:
    ticker: str
    model: Decimal
    comparison: Decimal
    actual: Money

    @property
    def diff(self):
        return self.model - Decimal(self.comparison)


def compare_portfolios(
    real: PortfolioProtocol,
    ideal: IdealPortfolio,
    buy_order=PurchaseStrategy.LARGEST_DIFF_FIRST,
    target_size: Optional[Union[Decimal, int]] = None,
):
    output: Dict[str, ComparisonResult] = {}
    diff = Decimal(0.0)
    selling = Decimal(0.0)
    buying = Decimal(0.0)
    target_value: Money = (
        Money(value=Decimal(target_size)) if target_size else real.value
    )
    for value in ideal.holdings:
        comparison = real.get_holding(value.ticker)
        if not comparison:
            percentage = Decimal(0.0)
            actual_value = Money.parse("0.0")
        else:
            percentage = Decimal((comparison.value / target_value).value)
            actual_value = comparison.value
        output[value.ticker] = ComparisonResult(
            ticker=value.ticker,
            model=value.weight,
            comparison=percentage,
            actual=actual_value,
        )
        _diff = Decimal(value.weight) - percentage
        diff += abs(_diff)
        if _diff < 0:
            selling += abs(_diff)
        else:
            buying += abs(_diff)

    Logger.info(
        f"Total portfolio % delta {print_per(diff)}. Overweight {print_per(selling)}, underweight {print_per(buying)}"
    )
    if buy_order == PurchaseStrategy.LARGEST_DIFF_FIRST:
        diff_output: Dict[str, ComparisonResult] = {
            k: v for k, v in sorted(output.items(), key=lambda item: -abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.CHEAPEST_FIRST:
        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    else:
        raise ValueError("Invalid purchase strategy")
    to_purchase = {}
    to_sell = {}
    for key, diffvalue in diff_output.items():
        if diffvalue.diff == 0:
            continue
        elif diffvalue.diff < 0:
            diff_text = "Overweight"
            to_sell[key] = (
                target_value * diffvalue.comparison - target_value * diffvalue.model
            )
        else:
            diff_text = "Underweight"
            to_purchase[key] = (
                target_value * diffvalue.model - target_value * diffvalue.comparison
            )
            # to_purchase.append(key)
        Logger.info(
            f"{diff_text} {key}, {print_per(diffvalue.model)} target vs {print_per(diffvalue.comparison)} actual. Should be {target_value * diffvalue.model}, is {diffvalue.actual}"
        )
    return to_purchase, to_sell


def round_with_strategy(to_buy_currency, rounding_strategy: RoundingStrategy) -> Money:
    if rounding_strategy == RoundingStrategy.CLOSEST:
        to_buy_units = Money(value=int(round(to_buy_currency, 0)))
    elif rounding_strategy == RoundingStrategy.FLOOR:
        to_buy_units = Money(value=floor(to_buy_currency))
    elif rounding_strategy == RoundingStrategy.CEILING:
        to_buy_units = Money(value=ceil(to_buy_currency))
    else:
        raise ValueError("Invalid Rounding Strategy")
    return to_buy_units


def generate_auto_target_size(
    real: CompositePortfolio,
    ideal: IdealPortfolio,
) -> Money:
    cash = Money(value=0)
    for input in real.portfolios:
        cash += input.cash
    in_portfolio_value = Money(value=0)
    for value in ideal.holdings:
        comparison = real.get_holding(value.ticker)
        if not comparison:
            continue
        else:
            in_portfolio_value += comparison.value
    return in_portfolio_value + cash


def generate_composite_order_plan(
    composite: CompositePortfolio,
    ideal: IdealPortfolio,
    purchase_order_maps: Mapping[Provider, PurchaseStrategy] | PurchaseStrategy,
    # rounding_strategy=RoundingStrategy.CLOSEST,
    purchase_power: Optional[Money | float | int] = None,
    target_size: Optional[Money | float | int] = None,
    min_order_value: Money = MIN_ORDER_MONEY,
    safety_threshold: Decimal = Decimal(0.95),
) -> Mapping[Provider, OrderPlan]:
    provider_map = {x.provider: x for x in composite.portfolios if x.provider}
    providers: List[BaseProvider] = list(provider_map.keys())  # type: ignore

    if isinstance(purchase_order_maps, PurchaseStrategy):
        purchase_order_maps = {x.PROVIDER: purchase_order_maps for x in providers}
    processed = set()
    # check each of our p
    output: defaultdict[Provider, OrderPlan] = defaultdict(
        lambda: OrderPlan(to_buy=[], to_sell=[])
    )
    skip_tickers: set[str] = set()
    for provider in providers:
        skip_tickers = skip_tickers.union(provider.get_unsettled_instruments())

    purchase_power_money = Money(value=purchase_power) if purchase_power else None

    while providers:
        provider = providers.pop()
        if purchase_power_money and purchase_power_money < Money(value=0):
            Logger.info("No dollars left to purchase")
            continue
        Logger.info(f"Beginning plan for {provider}")
        processed.add(provider.PROVIDER)
        port = provider_map[provider]
        # build the plan across the _entire_ composite portfolio

        # if we don't know how much cash we have, skip
        print(f"doing provider {provider} with {port.cash}")
        if not port.cash:
            continue

        local_max_spend = port.cash * safety_threshold
        if purchase_power_money:
            local_purchase_power = min(
                purchase_power_money * safety_threshold, local_max_spend
            )
            purchase_power_money = purchase_power_money - local_purchase_power
        else:
            local_purchase_power = port.cash * safety_threshold

        purchase_plan = generate_order_plan(
            ideal=ideal,
            real=composite,
            buy_order=purchase_order_maps[provider.PROVIDER],
            # rounding_strategy=RoundingStrategy.CLOSEST,
            target_size=target_size,
            purchase_power=local_purchase_power,
            min_order_value=min_order_value,
            skip_tickers=skip_tickers,
        )
        for ticker in purchase_plan.tickers:
            skip_tickers.add(ticker)
        output[provider.PROVIDER] += purchase_plan
    return output


def generate_order_plan(
    real: PortfolioProtocol,
    ideal: IdealPortfolio,
    buy_order=PurchaseStrategy.LARGEST_DIFF_FIRST,
    # rounding_strategy=RoundingStrategy.CLOSEST,
    target_size: Optional[Money | float | int] = None,
    purchase_power: Optional[Money | float | int] = None,
    min_order_value: Money = MIN_ORDER_MONEY,
    skip_tickers: Optional[set[str]] = None
    # fractional_shares: bool = True,
) -> OrderPlan:
    diff = Decimal(0.0)
    selling = Decimal(0.0)
    buying = Decimal(0.0)
    target_value: Money = Money(value=target_size) if target_size else real.value
    output: Dict[str, ComparisonResult] = {}
    purchase_power = Money(value=purchase_power or target_value)
    currently_held = Money(value=0)
    for value in ideal.holdings:
        if skip_tickers and value.ticker in skip_tickers:
            continue
        comparison = real.get_holding(value.ticker)
        if not comparison:
            percentage = Decimal(0.0)
            actual_value = Money.parse("0.0")
        else:
            percentage = Decimal((comparison.value / target_value).value)
            actual_value = comparison.value

        # track how much we currently have
        currently_held += actual_value
        output[value.ticker] = ComparisonResult(
            ticker=value.ticker,
            model=value.weight,
            comparison=percentage,
            actual=actual_value,
        )
        _diff = Decimal(value.weight) - percentage
        diff += round(abs(_diff), 4)
        if _diff == 0:
            continue
        elif _diff < 0:
            selling += abs(_diff)
        else:
            buying += abs(_diff)

    Logger.info(
        f"Total portfolio % delta {print_per(diff)}. Overweight {print_per(selling)}, underweight {print_per(buying)}"
    )

    scaling_factor = Money(value=1.0)

    if buy_order == PurchaseStrategy.LARGEST_DIFF_FIRST:
        diff_output: Dict[str, ComparisonResult] = {
            k: v for k, v in sorted(output.items(), key=lambda item: -abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.CHEAPEST_FIRST:
        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    elif buy_order == PurchaseStrategy.PEANUT_BUTTER:
        # divide the difference between where we want to be
        # and where we are
        # across all stocks
        scaling_factor = purchase_power / (target_value - currently_held)

        diff_output = {
            k: v for k, v in sorted(output.items(), key=lambda item: abs(item[1].diff))
        }
    else:
        raise ValueError("Invalid purchase strategy")
    to_purchase: list[OrderElement] = []
    to_sell: list[OrderElement] = []

    # first sell everything
    for key, diffvalue in diff_output.items():
        if round(diffvalue.diff, 4) == 0.0000:
            continue
        elif diffvalue.diff < 0:
            diff_text = "Overweight"
            sell_target: Money = (
                target_value * diffvalue.comparison - target_value * diffvalue.model
            )
            if buy_order == PurchaseStrategy.PEANUT_BUTTER:
                sell_target = sell_target * scaling_factor
            # if not fractional_shares:
            #     price = real.get_instrument_price(key)
            #     qty = round_with_strategy(target/price, rounding_strategy)
            #     target =
            sell_target = max(sell_target, MIN_ORDER_MONEY)
            to_sell.append(
                OrderElement(
                    ticker=key, value=sell_target, order_type=OrderType.SELL, qty=None
                )
            )
            #     target_value * diffvalue.comparison - target_value * diffvalue.model
            # )

    for key, diffvalue in diff_output.items():
        if purchase_power <= 0:
            break
        elif round(diffvalue.diff, 4) == 0.0000:
            continue
        elif diffvalue.diff > 0:
            diff_text = "Underweight"
            buy_target: Money = Money(
                value=min(
                    target_value * diffvalue.model
                    - target_value * diffvalue.comparison,
                    purchase_power,
                )
            )
            if buy_order == PurchaseStrategy.PEANUT_BUTTER:
                if buy_target > 0.0:
                    max_value: Decimal = max(
                        Decimal(float(buy_target.value)) * scaling_factor.decimal,
                        Decimal(1.0),
                    )
                    buy_target = Money(value=max_value)
            buy_target = max(buy_target, min_order_value)
            to_purchase.append(
                OrderElement(
                    ticker=key, value=buy_target, qty=None, order_type=OrderType.BUY
                )
            )
            purchase_power = purchase_power - buy_target

        Logger.debug(
            f"{diff_text} {key}, {print_per(diffvalue.model)} target vs {print_per(diffvalue.comparison)} actual. Should be {target_value * diffvalue.model}, is {diffvalue.actual}"
        )

    return OrderPlan(to_buy=to_purchase, to_sell=to_sell)



from typing import Set
import re
from datetime import date
from decimal import Decimal
from pydantic import BaseModel, Field
from pathlib import Path
import json

from py_portfolio_index.models import IdealPortfolioElement, IdealPortfolio

QUARTER_TO_MONTH = {1: 1, 2: 4, 3: 7, 4: 10}


def parse_date_from_name(input: str) -> date | None:
    components = input.lower().split("_")
    year = None
    quarter = None
    for item in components:
        if re.match("[0-9]{4}", item):
            year = int(item)
        if re.match("q[1-4]", item):
            quarter = int(item[1])
    if not year or not quarter:
        return None

    return date(year=year, month=QUARTER_TO_MONTH[quarter], day=1)


class IndexInventory(BaseModel):
    csv_keys: Set[str] = Field(exclude=True)
    json_keys: Set[str] = Field(exclude=True)
    base: Path = Field(exclude=True)
    loaded: dict[str, IdealPortfolio] = Field(default_factory=dict)

    @property
    def keys(self) -> Set[str]:
        return self.csv_keys.union(self.json_keys)

    @classmethod
    def from_path(cls, path):
        path = Path(path)
        if path.is_file():
            path = path.parent
        json_keys = set()
        csv_keys = set()
        for f in path.iterdir():
            try:
                if f.suffix == ".csv":
                    csv_keys.add(f.stem)
                if f.suffix == ".json":
                    json_keys.add(f.stem)
            except FileNotFoundError:
                pass
        return IndexInventory(json_keys=json_keys, csv_keys=csv_keys, base=path)

    def __getitem__(self, item: str) -> IdealPortfolio:
        if item in self.keys:
            values = self.loaded.get(item, None)
            if values:
                return values
            values = self.get_values(item)
            self.loaded[item] = values
            return values
        raise KeyError(item)

    def get_values(self, item: str) -> IdealPortfolio:
        out = []
        start_date = None
        if item in self.json_keys:
            with open(self.base / f"{item}.json") as f:
                parsed = json.loads(f.read())
                start_date = date.fromisoformat(parsed["as_of"])
                for row in parsed["components"]:
                    out.append(
                        IdealPortfolioElement(
                            ticker=row["ticker"], weight=Decimal(row["weight"])
                        )
                    )
        elif item in self.csv_keys:
            with open(self.base / f"{item}.csv") as f:
                contents = f.read()
                for row in contents.split("\n"):
                    ticker, weight = row.split(",", 1)
                    start_date = parse_date_from_name(item)
                    out.append(
                        IdealPortfolioElement(ticker=ticker, weight=Decimal(weight))
                    )
        else:
            raise ValueError("No matching file {}".format(item))

        if start_date:
            return IdealPortfolio(holdings=out, source_date=start_date)
        return IdealPortfolio(holdings=out)
